// Diese Programm steht unter der GPL
// (c) Patrick Kilian 2004-2005

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/time.h>
#include <SDL.h>

#define PI 3.121592654
#define DEBUG 0				// 1: Keine Bild-Ausgabe
#define VERBOSE 0			// 1: Statusmeldungen nach stderr
#define TIMEING 0			// 1: Timingmeldungen nach stderr

unsigned int hD;			// sichtbarer horizontaler Bereich in px
unsigned int vD;			// sichtbarer vertikaler Bereich in px
unsigned int hT;			// horizontaler Bereich in px
unsigned int vT;			// vertikaler Bereich in px
unsigned int pC;			// Pixelclock
unsigned int rF;			// Traegerfrequenz
unsigned int periodeT;		// Periodenlaenge des Traegers in px
unsigned int lineOffset;	// Offset pro Zeile durch Zeilenrücklauf
unsigned int zeilenFreq;	// Zeilenfrequenz in Hz
unsigned int repaintFreq;	// Bildwiederholfrequenz in Hz
SDL_Surface *screen;		// Bildschirmflaeche
SDL_Rect bg;				// Malfläche
SDL_Rect point;				// einzelnes Pixel
SDL_Event event;			// Maus, Tastatur, etc.
SDL_Color col[256];			// Farbpalette
unsigned int scancode;		// Keycode des eingelesen Zeichens
unsigned int pal[256];		// Farbpalette von schwarz bis weiss
SDL_Surface * ton[17];		// Vorberechnete Toene

/*
 * Liefert Monitorbild fuer einen Ton
 * freq Tonhoehe in Hz
 * freq == -1 bedeutet schwarzer Bildschirm
 */
SDL_Surface* create(int freq) {
	unsigned int periodeN = zeilenFreq / freq;	// Periodenlaenge in px
	unsigned int offset = 0;		// in px
	unsigned int strengthM; 		// Amplitude des aufmodulierten Signals
	unsigned int strengthT;			// Amplitude des Traegersignals
	unsigned int strengthF;			// Amplitude des fertigen Signals
	unsigned int traeger[256];		// vorberechnete Farbskala fuer diese Zeile
	double omegaT = 2 * PI / periodeT;
	// Bildschirmflaeche
	SDL_Surface* surface = SDL_CreateRGBSurface(SDL_HWSURFACE, hD, vD, 8, 0xff000000, 0x00ff0000, 0x0000ff00, 0x00000000);
	if(surface == NULL) {
		fprintf(stderr, "CreateRGBSurface failed: %s\n", SDL_GetError());
		exit(1);
	}

	SDL_SetColors(surface, col, 0, 256);
	for (int j = 0; j < 256; j++) {
		pal[j] = SDL_MapRGB(surface->format, j, j, j);
	}

	SDL_FillRect(surface, &bg, pal[0]);			// schwarzer Hintergrund

	if(freq == -1) {
		SDL_UpdateRect(surface, 0, 0, hD, vD);
		// return;
	}

	if ((freq < 50) || (freq > 20000)) {
		if (VERBOSE) {
			fprintf(stderr, "Unhörbarer Ton bei %d Hz.\n", freq);
		}
	}

	for (unsigned int zeile = 0; zeile < vD; zeile++) {
		strengthM = 8 * (sin(2 * PI * zeile / periodeN) + 1);
		// neue Zeile bedeutet Verschiebung um lineOffset
		offset = (offset  + lineOffset) % periodeT;
		if (strengthM > 0) {
			// Berechne Farbskala
			for (unsigned int i = 0; i < periodeT; i++) {
				strengthT = 8 * (sin(i * omegaT) + 1);
				strengthF = strengthT * strengthM;
				if (strengthF > 255) {
					strengthF = 255;
				}
				traeger[i] = strengthF;
			}
			// Zeichne Zeile
			for (unsigned int spalte = 0; spalte < hD; spalte++) {
				point.x = spalte;
				point.y = zeile;
				SDL_FillRect(surface, &point, pal[traeger[(spalte + offset)%periodeT]]);
			}
		}
	}
	return surface;
}

/*
 * Zeichnet das Monitorbild fuer einen Ton
 */
void play(SDL_Surface * bild) {
#if TIMEING
	struct timeval start;
	gettimeofday (&start, NULL);
#endif

	// Mache neuen Ton sichtbar
	if (SDL_BlitSurface(bild, NULL, screen, NULL) < 0) {
		fprintf(stderr, "Blit failed 4\n");
	}
	SDL_UpdateRect(screen, 0, 0, hD, vD);

#if TIMEING
	struct timeval end;
	gettimeofday (&end, NULL);

	int diffS = (end.tv_sec - start.tv_sec);
	int diffU = (end.tv_usec - start.tv_usec);
	int diff = (1000000 * diffS + diffU) / 1000;
	fprintf(stderr, "Paint braucht %d Millisec.\n", diff);
#endif
}
/*
 * Gibt einen Hilfe zu den Aufruf-Parametern
 */
void usage() {
	fprintf(stdout, "Aufruf-Parameter: \n");
	fprintf(stdout, "Benötigt werden (in dieser Reihenfolge):\n");
	fprintf(stdout, "pixelClock (in px/s)\n");
	fprintf(stdout, "hDisplay   (in px)\n");
	fprintf(stdout, "vDisplay   (in px)\n");
	fprintf(stdout, "hTotal     (in px)\n");
	fprintf(stdout, "vTotal     (in px)\n");
	fprintf(stdout, "radioFreq  (in 1/s)\n");
}

/*
 * main-Funktion
 * argc Anzahl der Aufrufparameter
 * args Array von char-Pointern die auf die Anfaenge der Parameter zeigen.
 * In args[0] steht der Programmname
 */
int main (int argc, char* args[]) {
	if (argc != 7) {
		usage();
		exit(2);
	}
	///////////////////////////////////////////////////////////////////////////
	// Pruefe und speichere Werte                                            //
	///////////////////////////////////////////////////////////////////////////

	// Pixelclock in px/s
	pC = strtoul(args[1], NULL, 10);
	if (pC < 20000000) {
		fprintf(stderr, "zu kleiner Pixeltakt %d\n", pC);
		exit(1);
	} else if ((pC < 100000000) || (pC > 200000000)) {
		fprintf(stderr, "ungewöhnlicher Pixeltakt %d\n", pC);
	} else {
		if (VERBOSE) {
			fprintf(stderr, "pC = %d\n", pC);
		}
	}

	// Breite des sichtbaren Bildschirmes in px
	hD = strtoul(args[2], NULL, 10);
	if (hD < 640) {
		fprintf(stderr, "zu kleine Bildbreite %d\n", hD);
		exit(1);
	} else if ((hD < 800) || (hD > 2000)) {
		if (VERBOSE) {
			fprintf(stderr, "ungewöhnliche Bildbreite %d\n", hD);
		}
	} else {
		if (VERBOSE) {
			fprintf(stderr, "hD = %d\n", hD);
		}
	}

	// Hoehe des sichtbaren Bildschirmes in px
	vD = strtoul(args[3], NULL, 10);
	if (vD < 480) {
		fprintf(stderr, "zu kleine Bildhöhe %d\n", vD);
		exit(1);
	} else if ((vD < 600) || (vD > 1500)) {
		fprintf(stderr, "ungewöhnliche Bildhöhe %d\n", vD);
	} else {
		if (VERBOSE) {
			fprintf(stderr, "vD = %d\n", vD);
		}
	}

	// Breite des Bildschirmes in px
	hT = strtoul(args[4], NULL, 10);
	if (hT < 640) {
		fprintf(stderr, "zu kleine Bildbreite %d\n", hT);
		exit(1);
	} else if (hT <= hD) {
		fprintf(stderr, "sehr kurzer Zeilenrücklaufes %d\n", hT - hD);
	} else {
		if (VERBOSE) {
			fprintf(stderr, "hT = %d\n", hT);
		}
	}


	// Hoehe des Bildschirmes in px
	vT = strtoul(args[5], NULL, 10);
	if (vT < 480) {
		fprintf(stderr, "zu kleine Bildhöhe %d\n", vT);
		exit(1);
	} else if (vT <= vD) {
		fprintf(stderr, "sehr kurzer Bildrücklaufes %d\n", vT - vD);
	} else {
		if (VERBOSE) {
			fprintf(stderr, "vT = %d\n", vT);
		}
	}

	// Radiofrequenz in Hz
	rF = strtoul(args[6], NULL, 10);
	if (rF < 6000000) {
		fprintf(stderr, "zu niedrige Trägerfrequenz %d\n", rF);
		exit(1);
	} else if ((rF < 9000000) || (rF > 11000000)) {
		fprintf(stderr,
			"ungewöhnliche Trägerfreuenz (nicht im 10MHz-Band) %d\n", rF);
	} else {
		if (VERBOSE) {
			fprintf(stderr, "rF = %d\n", rF);
		}
	}

	// Abgeleitete Groessen

	periodeT = pC / rF;
	lineOffset = hT % periodeT;
	if (VERBOSE) {
		fprintf(stderr, "lineOffset = %d\n", lineOffset);
	}
	zeilenFreq = pC / hT;
	repaintFreq = pC / (hT * vT);

	///////////////////////////////////////////////////////////////////////////
	// Video-Init                                                            //
	///////////////////////////////////////////////////////////////////////////
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Kann SDL nicht starten.\n(%s)\n", SDL_GetError());
		exit(1);
	}

#if DEBUG
	screen = SDL_SetVideoMode(100, 100, 8, SDL_HWSURFACE | SDL_ANYFORMAT);
# else
	screen = SDL_SetVideoMode(hD, vD, 8, SDL_HWSURFACE | SDL_ANYFORMAT | SDL_FULLSCREEN);
#endif

	if ( screen == NULL ) {
		fprintf(stderr, "Konnte den Videomodus nicht setzen.\n(%s)\n", SDL_GetError());
		exit(1);
	}

	if (VERBOSE) {
		fprintf(stdout, "%d bits-per pixel\n", screen->format->BitsPerPixel);
	}

	for (int i = 0; i < 256; i++) {
		col[i].r = i;
		col[i].g = i;
		col[i].b = i;
	}

	SDL_SetColors(screen, col, 0, 256);
	for (int j = 0; j < 256; j++) {
		pal[j] = SDL_MapRGB(screen->format, j, j, j);
	}

	// Setzte Werte der Malflaeche
	bg.x = 0;
	bg.y = 0;
	bg.w = hD;
	bg.h = vD;

	point.w = 1;
	point.h = 1;

	///////////////////////////////////////////////////////////////////////////
	// Berechne Toene                                                        //
	///////////////////////////////////////////////////////////////////////////
	ton[0] = create(-1);	// Stille
	ton[1] = create(440);	// A
	ton[2] = create(466);	// Ais
	ton[3] = create(494);	// H
	ton[4] = create(523);	// C
	ton[5] = create(554);	// Cis
	ton[6] = create(587);	// D
	ton[7] = create(622);	// Dis
	ton[8] = create(659);	// E
	ton[9] = create(698);	// F
	ton[10] = create(740);	//Fis
	ton[11] = create(784);	// G
	ton[12] = create(831);	// Gis
	ton[13] = create(880);	// A
	ton[14] = create(932);	// Ais
	ton[15] = create(988);	// H
	ton[16] = create(1047);	// C

	///////////////////////////////////////////////////////////////////////////
	// Event-Loop                                                            //
	///////////////////////////////////////////////////////////////////////////
	while (1) {
		SDL_WaitEvent(&event);
		switch (event.type) {
			case SDL_KEYDOWN: // Tastendruck
				scancode = event.key.keysym.scancode;
				switch (scancode) {
					case 38: //a
						if (VERBOSE) { fprintf(stdout, "A\n"); }
						if (! DEBUG) { play(ton[1]); }
						break;
					case 25: // w
						if (VERBOSE) { fprintf(stdout, "Ais\n"); }
						if (! DEBUG) { play(ton[2]); }
						break;
					case 39: // s
						if (VERBOSE) { fprintf(stdout, "H\n"); }
						if (! DEBUG) { play(ton[3]); }
						break;
					case 40: // d
						if (VERBOSE) { fprintf(stdout, "C\n"); }
						if (! DEBUG) { play(ton[4]); }
						break;
					case 27: // r
						if (VERBOSE) { fprintf(stdout, "Cis\n"); }
						if (! DEBUG) { play(ton[5]); }
						break;
					case 41: // f
						if (VERBOSE) { fprintf(stdout, "D\n"); }
						if (! DEBUG) { play(ton[6]); }
						break;
					case 28: // t
						if (VERBOSE) { fprintf(stdout, "Dis\n"); }
						if (! DEBUG) { play(ton[7]); }
						break;
					case 42: // g
						if (VERBOSE) { fprintf(stdout, "E\n"); }
						if (! DEBUG) { play(ton[8]); }
						break;
					case 43: // h
						if (VERBOSE) { fprintf(stdout, "F\n"); }
						if (! DEBUG) { play(ton[9]); }
						break;
					case 30: // u
						if (VERBOSE) { fprintf(stdout, "Fis\n"); }
						if (! DEBUG) { play(ton[10]); }
						break;
					case 44: // j
						if (VERBOSE) { fprintf(stdout, "G\n"); }
						if (! DEBUG) { play(ton[11]); }
						break;
					case 31: // i
						if (VERBOSE) { fprintf(stdout, "Gis\n"); }
						if (! DEBUG) { play(ton[12]); }
						break;
					case 45: // k
						if (VERBOSE) { fprintf(stdout, "A\n"); }
						if (! DEBUG) { play(ton[13]); }
						break;
					case 32: // o
						if (VERBOSE) { fprintf(stdout, "Ais\n"); }
						if (! DEBUG) { play(ton[14]); }
						break;
					case 46: // l
						if (VERBOSE) { fprintf(stdout, "H\n"); }
						if (! DEBUG) { play(ton[15]); }
						break;
					case 47: // ö
						if (VERBOSE) { fprintf(stdout, "C\n"); }
						if (! DEBUG) { play(ton[16]); }
						break;
					case 9: // ESC
						exit(0);
					default:
						if (DEBUG) { fprintf(stdout, "%d\n", (scancode)); }
						break;
				}
				break;
			case SDL_KEYUP: // Taste wird losgelassen
				play(ton[0]);
				break;
			case SDL_QUIT:  // Fenster wird geschlossen
				exit(0);
		}
	}
}
